import logging
from collections import defaultdict

import hydra
import torch
from omegaconf import DictConfig

from empiricalprivacyestimation.data import get_dataloaders
from empiricalprivacyestimation.epsilon_estimation import get_privacy_auditing_steps
from empiricalprivacyestimation.epsilon_estimation.util import get_privacy_profile
from empiricalprivacyestimation.models import get_model
from empiricalprivacyestimation.training import get_attack_scores
from empiricalprivacyestimation.util.cache_results import cache_results
from empiricalprivacyestimation.util.plot_histogram import plot_histogram
from empiricalprivacyestimation.util.plot_per_step import per_step_plot
from empiricalprivacyestimation.util.plot_privacy_profile import plot_privacy_profile


@hydra.main(version_base=None, config_path="conf", config_name="audit_each_param")
def main(cfg: DictConfig):
    out_dir = hydra.core.hydra_config.HydraConfig.get()["runtime"]["output_dir"]
    normal_loader, test_loader, attack_point = get_dataloaders(cfg)
    initialization = get_model(cfg)

    # get number of parameters
    num_parameters = sum(p.numel() for p in initialization.parameters() if p.requires_grad)
    cache_results(
        "initial_state_dict", {k: v.cpu() for k, v in initialization.state_dict().items()}
    )

    for i in range(num_parameters):
        results = defaultdict(list)

        cfg.attack.idx = i
        (
            initial_state_dict,
            attack_outputs_per_run,
            gt,
            epsilons,
            test_scores,
            sensitivity
        ) = get_attack_scores(
            cfg=cfg,
            normal_loader=normal_loader,
            test_loader=test_loader,
            attack_point=attack_point,
            state_dict=initialization.state_dict(),
        )

        privacy_auditing_steps = get_privacy_auditing_steps(cfg)
        audited_eps_values = []

        cache_results(f"attack_outputs_per_run_{i}", attack_outputs_per_run)
        logging.info(attack_outputs_per_run)

        cache_results(f"ground_truth_{i}", gt)
        cache_results(f"epsilons_{i}", epsilons)
        cache_results(f"sensitivity_{i}", sensitivity)

        alphas_per_run, betas_per_run = [], []


        for step in privacy_auditing_steps:
            attack_outputs = attack_outputs_per_run[:, step]
            if cfg.privacy.compute_epsilons:
                audited_eps_values.append(epsilons[step])
            alphas, betas, empirical_epsilons_cp, empirical_epsilons_gdp = get_privacy_profile(
                cfg, gt=gt, outputs=attack_outputs
            )

            alphas_per_run.append(alphas)
            betas_per_run.append(betas)

            results[f"clopper-pearson_lb"].append(max(empirical_epsilons_cp))
            results[f"gdp-cp_lb"].append(max(empirical_epsilons_gdp))
            in_samples = []
            out_samples = []
            for (param, label) in zip(attack_outputs, gt):
                if label == 0:
                    in_samples.append(param)
                else:
                    out_samples.append(param)

            if cfg.logging_params.log_every_step:
                plot_histogram(
                    filename=f"{out_dir}/param_distribution_step_{step}.pdf",
                    in_samples=torch.tensor(in_samples),
                    out_samples=torch.tensor(out_samples),
                    in_label="dataset with anomaly",
                    out_label="dataset without anomaly",
                )

                plot_privacy_profile(
                    filename=f"{out_dir}/privacy_profile_step_{step}.pdf",
                    alphas=alphas,
                    betas=betas,
                )

        cache_results(f"auditing_results_{i}", results)
        cache_results(f"alphas_per_run_{i}", alphas_per_run)
        cache_results(f"betas_per_run_{i}", betas_per_run)

        per_step_plot(
            cfg=cfg,
            steps=privacy_auditing_steps,
            results=results,
            th_eps=audited_eps_values,
            path=out_dir,
            title=f"{cfg.model.name} per step performance_{i}",
        )


if __name__ == "__main__":
    main()
