import logging

import hydra
from omegaconf import DictConfig

from empiricalprivacyestimation.data import get_dataloaders
from empiricalprivacyestimation.training import get_attack_scores
from empiricalprivacyestimation.util.cache_results import cache_results
from empiricalprivacyestimation.util.curve_confidence_interval import plot_CI_curves


@hydra.main(version_base=None, config_path="conf", config_name="sensitivity_analysis")
def main(cfg: DictConfig):
    samples = []
    for i in range(cfg.number_of_samples):
        logging.info(f"Run {i}")
        normal_loader, test_loader, attack_point = get_dataloaders(cfg)
        (
            initial_state_dict,
            attack_outputs_per_run,
            gt,
            epsilons,
            test_scores,
            sensitivity,
        ) = get_attack_scores(
            cfg=cfg,
            normal_loader=normal_loader,
            test_loader=test_loader,
            attack_point=attack_point,
        )
        cache_results(
            f"initial_state_dict_{i}",
            {k: v.cpu() for k, v in initial_state_dict.items()},
        )
        cache_results(f"attack_outputs_per_run_{i}", attack_outputs_per_run)
        cache_results(f"ground_truth_{i}", gt)
        cache_results(f"epsilons_{i}", epsilons)
        cache_results(f"test_scores_{i}", test_scores)
        samples.append(sensitivity)

    cache_results("samples", samples)
    plot_CI_curves(
        samples=samples,
        xlabel="Step",
        ylabel="Sensitivity",
        filename="sensitivity_analysis.pdf",
        C=cfg.privacy.C,
    )


if __name__ == "__main__":
    main()
