import logging
from typing import Tuple, List

import hydra
import numpy as np
from omegaconf import DictConfig
from scipy.stats import norm
from tqdm import tqdm

from empiricalprivacyestimation.epsilon_estimation.util import get_privacy_profile
from empiricalprivacyestimation.util.cache_results import cache_results


def get_gt(alpha, C, sigma):
    return norm.cdf(norm.ppf(1 - alpha) - C/sigma)


def gradient(x: float, C: float) -> float:
    if x < C/2:
        return -C
    return C


def gradient_descent(iterations: int, sigma: float, C: float, k: int) -> Tuple[List[float], int]:
    b = np.random.randint(0, 2, size=1)
    xs = []
    x = 0

    for i in range(iterations):
        if i % k == 0:
            derivative = 0
        else:
            derivative = gradient(x, C=C)

        if b == 0 and i % k == 0:
            x = x + (derivative + np.random.normal(loc=C, scale=sigma))
        else:
            x = x + (derivative + np.random.normal(loc=0, scale=sigma))
        xs.append(x)

    return xs, b


@hydra.main(version_base=None, config_path="conf", config_name="syntethic_example")
def main(cfg: DictConfig):
    C = cfg.privacy.C
    sigma = cfg.privacy.sigma
    results = []

    for i in tqdm(range(5)):
        xs = []
        bs = []

        for _ in range(cfg.empirical_epsilon.attack_samples):
            x, b = gradient_descent(
                iterations=cfg.training_params.total_no_of_steps,
                sigma=sigma,
                C=C,
                k=cfg.training_params.k,
            )
            xs.append(x)
            bs.append(b)

        xs = np.array(xs)
        bs = np.array(bs)

        cache_results("xs", xs)
        cache_results("bs", bs)

        epsilons_privacy_gdp = []
        privacy_profiles_per_step = []

        logging.info("Computing privacy profile")
        for j in tqdm(range(cfg.training_params.total_no_of_steps)):
            # if j % cfg.training_params.k == 0:
            alphas, betas, eps_cp, eps_gdp = get_privacy_profile(
                cfg=cfg, gt=bs, outputs=xs[:, j], alpha=0.05
            )
            epsilons_privacy_gdp.append(max(eps_gdp))
            privacy_profiles_per_step.append((alphas, betas))
        epsilons_privacy_gdp = np.array(epsilons_privacy_gdp)
        cache_results(f"epsilons_privacy_gdp_{i}", epsilons_privacy_gdp)
        results.append((epsilons_privacy_gdp, privacy_profiles_per_step))
    cache_results("results", results)
#
if __name__ == "__main__":
    main()
