import logging
from typing import List, Tuple

import hydra
import numpy as np
from omegaconf import DictConfig
from scipy.stats import norm
from tqdm import tqdm

from empiricalprivacyestimation.epsilon_estimation.util import get_privacy_profile
from empiricalprivacyestimation.util.cache_results import cache_results


def get_gdp_tradeoff_function(alpha, sigma, C):
    return norm.cdf(norm.ppf(1 - alpha) - C/sigma)


def gradient(x: float, threshold_function, C: float) -> float:
    if threshold_function(x):
        return C
    return -C


def gradient_descent(sigma: float, threshold_function, C: float) -> Tuple[List[float], int]:
    x = 0
    b = np.random.randint(0, 2, size=1)
    xs = []

    for i in range(2):
        if i == 0:
            derivative = 0
        else:
            derivative = gradient(x, threshold_function, C=C)

        if b == 0 and i == 0:
            x = x + (derivative + np.random.normal(C, scale=sigma))
        else:
            x = x + (derivative + np.random.normal(0, scale=sigma))
        xs.append(x)

    return xs, b


@hydra.main(version_base=None, config_path="conf", config_name="syntethic_example")
def main(cfg: DictConfig):
    C = cfg.privacy.C

    for sigma in tqdm([1, 1.5, 2, 2.5, 3, 3.5, 4, 4.5, 5]):
        # discretize the GDP tradeoff function for the Gaussian mechanism
        fixed_point = norm.cdf(-C/(2*sigma), loc=0, scale=sigma)

        alphas = np.linspace(0 + cfg.privacy.delta, fixed_point, 20)
        betas = [get_gdp_tradeoff_function(alpha=alpha, sigma=sigma, C=C) for alpha in alphas]

        # for each pair find the associated threshold function
        thresholds = []
        for alpha, beta in zip(alphas, betas):
            threshold = norm.ppf(1 - alpha, loc=0, scale=sigma)
            thresholds.append(threshold)

        for rep in range(1):
            epsilons_privacy_gdp = []
            privacy_profiles = []
            for i, (alpha, beta, threshold) in enumerate(zip(alphas, betas, thresholds)):
                xs = []
                bs = []

                threshold_function = lambda x: x > threshold

                for _ in range(cfg.empirical_epsilon.attack_samples):
                    x, b = gradient_descent(
                        sigma=sigma,
                        threshold_function=threshold_function,
                        C=C,
                    )
                    xs.append(x)
                    bs.append(b)

                xs = np.array(xs).squeeze()
                bs = np.array(bs).squeeze()

                cache_results(f"xs_{i}_{rep}", xs)
                cache_results(f"bs_{i}_{rep}", bs)

                logging.info("Computing privacy profile")
                alphas, betas, eps_cp, eps_gdp = get_privacy_profile(
                    cfg=cfg, gt=bs, outputs=xs[:, 1], alpha=0.05
                )
                epsilons_privacy_gdp.append(max(eps_gdp))
                privacy_profiles.append((alphas, betas))

            logging.info(f"epsilons_privacy_gdp={epsilons_privacy_gdp}")
            epsilons_privacy_gdp = np.array(epsilons_privacy_gdp)
            cache_results(f"epsilons_privacy_gdp_{sigma}_{rep}", epsilons_privacy_gdp)
            cache_results(f"privacy_profiles_{sigma}_{rep}", privacy_profiles)



if __name__ == "__main__":
    main()
