import logging
from collections import defaultdict

import hydra
from omegaconf import DictConfig

from empiricalprivacyestimation.data import get_dataloaders
from empiricalprivacyestimation.epsilon_estimation import get_privacy_auditing_steps
from empiricalprivacyestimation.epsilon_estimation.util import get_privacy_profile
from empiricalprivacyestimation.models import get_model
from empiricalprivacyestimation.training import get_attack_scores
from empiricalprivacyestimation.util.cache_results import cache_results


@hydra.main(version_base=None, config_path="conf", config_name="hidden_state")
def main(cfg: DictConfig):
    normal_loader, test_loader, attack_point = get_dataloaders(cfg)
    results = defaultdict(list)
    initialization = get_model(cfg)

    (
        initial_state_dict,
        attack_outputs_per_run,
        gt,
        epsilons,
        test_scores,
        sensitivity
    ) = get_attack_scores(
        cfg=cfg,
        normal_loader=normal_loader,
        test_loader=test_loader,
        attack_point=attack_point,
        state_dict=initialization.state_dict(),
    )

    privacy_auditing_steps = get_privacy_auditing_steps(cfg)
    audited_eps_values = []

    cache_results(
        "initial_state_dict", {k: v.cpu() for k, v in initial_state_dict.items()}
    )
    cache_results("attack_outputs_per_run", attack_outputs_per_run)
    logging.info(attack_outputs_per_run)

    cache_results("ground_truth", gt)
    cache_results("epsilons", epsilons)
    cache_results("sensitivity", sensitivity)
    cache_results("test_scores", test_scores)

    alphas_per_run, betas_per_run = [], []


    for step in privacy_auditing_steps:
        attack_outputs = attack_outputs_per_run[:, step]
        if cfg.privacy.compute_epsilons:
            audited_eps_values.append(epsilons[step])
        alphas, betas, empirical_epsilons_cp, empirical_epsilons_gdp = get_privacy_profile(
            cfg, gt=gt, outputs=attack_outputs
        )

        alphas_per_run.append(alphas)
        betas_per_run.append(betas)

        results[f"clopper-pearson_lb"].append(max(empirical_epsilons_cp))
        results[f"gdp-cp_lb"].append(max(empirical_epsilons_gdp))
        in_samples = []
        out_samples = []
        for (param, label) in zip(attack_outputs, gt):
            if label == 0:
                in_samples.append(param)
            else:
                out_samples.append(param)

    cache_results("auditing_results", results)
    cache_results("alphas_per_run", alphas_per_run)
    cache_results("betas_per_run", betas_per_run)

if __name__ == "__main__":
    main()
