import logging
from collections import defaultdict

import hydra
from omegaconf import DictConfig

from empiricalprivacyestimation.data import get_dataloaders
from empiricalprivacyestimation.epsilon_estimation import get_privacy_auditing_steps
from empiricalprivacyestimation.epsilon_estimation.util import get_privacy_profile
from empiricalprivacyestimation.models import get_model
from empiricalprivacyestimation.training import get_attack_scores
from empiricalprivacyestimation.util.cache_results import cache_results


@hydra.main(version_base=None, config_path="conf", config_name="low_dim_experiment")
def main(cfg: DictConfig):
    normal_loader, test_loader, attack_point = get_dataloaders(cfg)
    initialization = get_model(cfg)
    auditing_results = defaultdict(list)

    attacks = [
        # ("random", None),
        ("noisy_training", "per_step_norm"),
        ("noisy_training", "final_model_diff"),
        ("noiseless_training", "per_step_norm"),
        ("noiseless_training", "final_model_diff")
    ]
    for attack_name, scoring in attacks:
        for i in range(cfg.number_of_runs):
            results = defaultdict(list)

            logging.info(f"Running attack {i} {attack_name} with scoring {scoring}")
            cfg.attack.parameter_index = None
            cfg.attack.parameter_index_selection = attack_name
            cfg.attack.score = scoring

            (
                initial_state_dict,
                attack_outputs_per_run,
                gt,
                epsilons,
                test_scores,
                sensitivity
            ) = get_attack_scores(
                cfg=cfg,
                normal_loader=normal_loader,
                test_loader=test_loader,
                attack_point=attack_point,
                state_dict=initialization.state_dict(),
            )
            logging.info(f"Test scores: {test_scores}")

            privacy_auditing_steps = get_privacy_auditing_steps(cfg)
            audited_eps_values = []

            cache_results(
                f"initial_state_dict_{attack_name}_{scoring}_{i}", {k: v.cpu() for k, v in initial_state_dict.items()}
            )
            cache_results(
                f"initial_state_dict_{attack_name}_{scoring}_{i}", {k: v.cpu() for k, v in initial_state_dict.items()}
            )
            cache_results(f"attack_outputs_per_run_{attack_name}_{scoring}_{i}", attack_outputs_per_run)
            logging.info(attack_outputs_per_run)

            cache_results(f"ground_truth_{attack_name}_{scoring}_{i}", gt)
            cache_results(f"epsilons_{attack_name}_{scoring}_{i}", epsilons)
            cache_results(f"sensitivity_{attack_name}_{scoring}_{i}", sensitivity)

            alphas_per_run, betas_per_run = [], []
            for step in privacy_auditing_steps:
                attack_outputs = attack_outputs_per_run[:, step]
                if cfg.privacy.compute_epsilons:
                    audited_eps_values.append(epsilons[step])
                alphas, betas, empirical_epsilons_cp, empirical_epsilons_gdp = get_privacy_profile(
                    cfg, gt=gt, outputs=attack_outputs
                )

                alphas_per_run.append(alphas)
                betas_per_run.append(betas)

                results[f"clopper-pearson_lb"].append(max(empirical_epsilons_cp))
                results[f"gdp-cp_lb"].append(max(empirical_epsilons_gdp))
                in_samples = []
                out_samples = []
                for (param, label) in zip(attack_outputs, gt):
                    if label == 0:
                        in_samples.append(param)
                    else:
                        out_samples.append(param)


            cache_results(f"auditing_results_{attack_name}_{scoring}_{i}", results)
            cache_results(f"alphas_per_run_{attack_name}_{scoring}_{i}", alphas_per_run)
            cache_results(f"betas_per_run_{attack_name}_{scoring}_{i}", betas_per_run)
            auditing_results[f"{attack_name}-{scoring}"].append(results[f"gdp-cp_lb"])

    cache_results(f"auditing_results", auditing_results)


if __name__ == "__main__":
    main()
