from collections import defaultdict
from functools import partial
from typing import Tuple

import hydra
import jax
import numpy as np
from jax import jit
from jax import numpy as jnp
from jax import random
from omegaconf import DictConfig
from tqdm import tqdm

from empiricalprivacyestimation.epsilon_estimation.util import get_privacy_profile
from empiricalprivacyestimation.util.cache_results import cache_results


@hydra.main(version_base=None, config_path="conf", config_name="syntethic_example")
def main(cfg: DictConfig):
    C = cfg.privacy.C
    sigmas = [1, 2, 3, 4, 5, 6, 7, 8]
    batch_sizes = [1, 2, 4, 8, 16]
    mapps = defaultdict(list)
    key = random.PRNGKey(0)  # Initialize the random seed

    for sigma in tqdm(sigmas):
        for batch_size in batch_sizes:
            for i in range(3):
                xs = []
                bs = []
                for _ in range(cfg.empirical_epsilon.attack_samples):
                    x, b, key = gradient_descent(
                        iterations=cfg.training_params.total_no_of_steps,
                        sigma=sigma,
                        C=C,
                        k=cfg.training_params.k,
                        batch_size=batch_size,
                        reject_threshold=C/2,
                        key=key
                    )
                    xs.append(x)
                    bs.append(b)
                xs = np.array(xs).squeeze()
                bs = np.array(bs).squeeze()
                for i in range(cfg.training_params.total_no_of_steps):
                    alphas, betas, eps_cp, eps_gdp = get_privacy_profile(
                        cfg=cfg, gt=bs, outputs=xs[:, i], alpha=0.05, threshold = 0.5
                    )
                    mapps[(sigma, i + 1, batch_size)].append(eps_gdp)
                    print(f"Sigma: {sigma} Batch size: {batch_size}, Step: {i + 1}, Emp_Eps: {eps_gdp}")

    cache_results(f"epsilons", mapps)


# JAX-compatible gradient descent function
@partial(jit, static_argnums=(0,))
def gradient_descent(iterations: int, sigma: float, C: float, k: int, batch_size: int, reject_threshold: float,
                     key: jnp.ndarray) -> Tuple[jnp.ndarray, int]:
    # Initialize random key
    key, subkey_b, prop = random.split(key, 3)

    # Generate random binary value b
    b = random.randint(subkey_b, shape=(), minval=0, maxval=2)

    # Initialize variables
    x = jnp.zeros(1, dtype=jnp.float32)
    derivative = jnp.zeros(1, dtype=jnp.float32)

    # Define the step function for the loop
    @jax.jit
    def step(carry, i):
        x, derivative, key = carry
        key, subkey_noise = random.split(key)

        # Add noise and update x
        noise = sigma * random.normal(subkey_noise, shape=(1,), dtype=jnp.float32)

        x = jax.lax.cond(
            (b == 0) & (i % k == 0),
            lambda _: x + C + derivative + noise,
            lambda _: x + derivative + noise,
            None
        )

        # Update derivative
        derivative = gradient(x, C=C, B=batch_size, reject_threshold=reject_threshold)

        return (x, derivative, key), x

    # Use `lax.scan` to loop over iterations and accumulate `xs`
    initial_carry = (x, derivative, key)
    _, xs = jax.lax.scan(step, initial_carry, jnp.arange(iterations))

    return xs, b, prop

# Vectorized gradient function using JAX
def gradient(x: jnp.ndarray, C: float, B: int, reject_threshold: float) -> jnp.ndarray:
    return jnp.where(x < C / 2, -B * C, B * C)

if __name__ == "__main__":
    main()
