import logging
from typing import Dict, Optional

from omegaconf import DictConfig
from torch.nn import Module
from torch.utils.data import DataLoader

from .canary_on_the_sphere import get_score, setup_canary_on_the_sphere_attack, generate_canary
from .loss_based_attack import loss_attack
from .poison import get_poison_attack_output, poison_attack_setup
from ..data.gaussian_data_outlier import DataPoint
from ..models import get_model

canary = None

def attack_setup(
    cfg: DictConfig,
    train_loader: DataLoader,
    test_loader: DataLoader,
    state_dict: Optional[Dict] = None,
) -> Dict:
    if cfg.attack.name == "loss":
        if state_dict:
            return state_dict

        model = get_model(cfg)
        state_dict = model.state_dict()
        return state_dict

    if cfg.attack.name == "poison":
        return poison_attack_setup(
            cfg=cfg,
            train_loader=train_loader,
            test_loader=test_loader,
            state_dict=state_dict,
        )

    if cfg.attack.name == "canary_on_the_sphere":
        setup_canary_on_the_sphere_attack(cfg)
        if state_dict:
            return state_dict
        model = get_model(cfg)
        return model.state_dict()

    raise ValueError(f"Attack {cfg.attack.name} not supported")


def get_attack_score(cfg: DictConfig, model: Module, attack_point: DataPoint, initialization: Module) -> float:
    if cfg.attack.name == "loss":
        return loss_attack(cfg, model, attack_point)

    if cfg.attack.name == "poison":
        return get_poison_attack_output(model, cfg.attack.parameter_index)

    if cfg.attack.name == "canary_on_the_sphere":
        global canary
        if canary is None:
            logging.error("Canary not initialized. Running the attack on the sphere")
            total_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
            canary = generate_canary(total_params).to("cuda")
            logging.error(f"Canary initialized successfully - {str(canary)}")

        return get_score(canary, model, initialization)

    raise ValueError(f"Attack {cfg.attack.name} not supported")
