import logging

import torch
from omegaconf import DictConfig
from torch.nn.utils import parameters_to_vector, vector_to_parameters

from ..models import get_model


def generate_canary(dimension):
    """
    Generates a deterministic random canary vector sampled uniformly from the unit sphere.
    """
    random_state = torch.get_rng_state()
    logging.info(random_state)
    torch.manual_seed(42)
    canary = torch.randn(dimension, device="cuda")
    canary /= torch.norm(canary)
    torch.set_rng_state(random_state)

    return canary

def compute_cosine_similarity(vec1, vec2):
    """
    Computes cosine similarity between two vectors.
    """
    dot_product = torch.dot(vec1, vec2)
    norm1 = torch.norm(vec1)
    norm2 = torch.norm(vec2)
    cosine_sim = dot_product / (norm1 * norm2)
    return cosine_sim.item()

@torch.compile
def get_score(canary, model, initialization):
    """
    Computes the score of the model based on the cosine similarity between
    the model updates and the canary vector.
    """
    model_params = parameters_to_vector([p for p in model.parameters() if p.requires_grad])
    init_params = parameters_to_vector([p for p in initialization.parameters() if p.requires_grad])

    # Compute the parameter update
    update_params = model_params - init_params

    # Compute cosine similarity directly
    score = compute_cosine_similarity(update_params, canary)
    return score


@torch.no_grad()
def setup_canary_on_the_sphere_attack(cfg: DictConfig):
    global canary
    """
    Generates and sets up the canary for the attack.
    """
    model = get_model(cfg)
    total_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    canary = generate_canary(total_params).to(cfg.training_params.device)

@torch.no_grad()
def embed_canary_in_model(cfg, model, reduction):
    from . import canary
    """
    Embeds the canary into the model parameters.
    """
    if reduction == "mean":
        magnitude = cfg.training_params.learning_rate * cfg.privacy.C / cfg.training_params.batch_size
    elif reduction == "sum":
        magnitude = cfg.training_params.learning_rate * cfg.privacy.C
    else:
        raise ValueError("Unknown loss reduction")

    # Flatten model parameters
    params = [p for p in model.parameters() if p.requires_grad]
    param_vector = parameters_to_vector(params)

    # Add scaled canary
    param_vector += canary * magnitude

    # Set the updated parameters back to the model
    vector_to_parameters(param_vector, params)