import logging
import random
from typing import Tuple, Dict, Optional

import numpy
import torch
from omegaconf import DictConfig
from opacus import PrivacyEngine
from torch import nn
from torch.utils.data import DataLoader

from empiricalprivacyestimation.models import get_model
from empiricalprivacyestimation.models.util import (
    no_params,
    flat_grad,
    param_shapes,
    flat_params,
)
from empiricalprivacyestimation.util.loss_fn import get_loss_fn
from empiricalprivacyestimation.util.optim import get_optim


def find_index(target_index: int, shapes: Tuple[Tuple, ...]) -> Tuple[int, ...]:
    """
    Find the index of a flattened parameter vector in the original parameter vector.
    :param target_index: the index of the flattened parameter vector
    :param shapes: the shapes of the unflattened parameters
    :return: the multidimensional index of the unflattened parameters
    """
    current_index = 0

    for layer_idx, shape in enumerate(shapes):
        new_index = 1
        for shape_dim in shape:
            new_index *= shape_dim

        new_index += current_index

        if current_index <= target_index < new_index:
            idx = numpy.unravel_index((target_index - current_index), tuple(shape))
            return layer_idx, *idx

        current_index = new_index

    raise ValueError


def extract_parameter(
    cfg: DictConfig,
    train_loader: DataLoader,
    test_loader: DataLoader,
    state_dict: Optional[Dict],
) -> Tuple[Dict, int, torch.Tensor]:
    """
    Run a training run to extract the gradient channel for poisoning based
    on a scoring. The training has to be made with the same particularities
    as the targeted training routine (e.g. same optimizer, same loss function, clipping)
    There are two possible ways of scoring:
    1. The minimum gradient norm of the gradient channel in time.
    2. The minimum absolute difference between initialization and the final model.

    The training can be:
    1. Noisy (multiple runs are required)
    2. Noiseless (only one run is required, as everything is deterministic)

    :param cfg: The config that describes the experiment
    :param train_loader: The training data loader
    :param test_loader: The test data loader
    :param state_dict: The state dict of the model from which to simulate the training
    :return: The state dict of the model, the index of the selected parameter, the gradient channel
        the scoring of all the points
    """
    from empiricalprivacyestimation.training.training_routines import test

    if state_dict is None:
        model = get_model(cfg)
        state_dict = model.state_dict()

    init_model = get_model(cfg)
    init_model.load_state_dict(state_dict)
    device = torch.device(cfg.training_params.device)

    logging.info("Starting to extract the min gradient channel for poisoning")
    score_accumulation = torch.zeros(no_params(init_model))

    for _ in range(cfg.attack.samples):
        sample_score = torch.zeros_like(score_accumulation)
        model = get_model(cfg)
        model.load_state_dict(state_dict)

        optimizer = get_optim(cfg, model=model)
        privacy_engine = PrivacyEngine()

        model, optimizer, poisson_train_loader = privacy_engine.make_private(
            module=model,
            optimizer=optimizer,
            data_loader=train_loader,
            noise_multiplier=cfg.privacy.sigma,
            max_grad_norm=cfg.privacy.C,
            loss_reduction="sum",
        )

        if cfg.training_params.poisson_sampled_dataset:
            train_loader = poisson_train_loader

        criterion, reduction = get_loss_fn(cfg)

        # TODO: GPU support
        for curr_step in range(1, cfg.training_params.total_no_of_steps + 1):
            model.train()

            try:
                data, target = next(dataloader_iterator)
            except:
                dataloader_iterator = iter(train_loader)
                data, target = next(dataloader_iterator)

            data, target = data.to(device), target.to(
                device
            )

            optimizer.zero_grad()
            output = model(data)
            loss = criterion(output, target)
            loss.backward()
            optimizer.step()

            with torch.no_grad():
                if cfg.attack.score == "per_step_norm":
                    with torch.no_grad():
                        sample_score += flat_grad(model).cpu() ** 2

        with torch.no_grad():
            test(cfg, model, test_loader)
            if cfg.attack.score == "per_step_norm":
                sample_score = torch.sqrt(sample_score).cpu()

            if cfg.attack.score == "final_model":
                sample_score = torch.abs(
                    flat_params(init_model) - flat_params(model)
                ).cpu()

            score_accumulation += sample_score

    with torch.no_grad():
        score_accumulation /= cfg.attack.samples
        parameter_index = torch.argmin(score_accumulation).cpu().item()

    return state_dict, parameter_index, score_accumulation


def poison_attack_setup(
    cfg: DictConfig,
    train_loader: DataLoader,
    test_loader: DataLoader,
    state_dict: Optional[Dict] = None,
) -> Dict:
    """
    Setup the poisoning attack. This includes:
    1. Extracting the target parameter
    2. Create a (possibly favorable for the attacker) initialization

    :param cfg: The hyperparameter of the experiment
    :param train_loader: The training data loader
    :param test_loader: The test data loader
    :param state_dict: (Optional) The state dict of the model from which to simulate the training

    :return: The state dict of the model
    """
    if cfg.attack.parameter_index_selection == "noiseless_training":

        old_sigma = cfg.privacy.sigma
        cfg.privacy.sigma = 0.0
        cfg.attack.samples = 1
        (
            state_dict,
            parameter_index,
            scores,
        ) = extract_parameter(cfg, train_loader, test_loader, state_dict)
        cfg.privacy.sigma = old_sigma
    elif cfg.attack.parameter_index_selection == "noisy_training":
        (
            state_dict,
            parameter_index,
            scores,
        ) = extract_parameter(cfg, train_loader, test_loader, state_dict)
    elif cfg.attack.parameter_index_selection == "random":
        model = get_model(cfg)

        parameter_index = random.randint(
            0, sum(p.numel() for p in model.parameters() if p.requires_grad) - 1
        )
        scores = None
    elif cfg.attack.parameter_index_selection == "each":
        parameter_index = cfg.attack.idx
        scores = None
    else:
        raise ValueError("Unknown target parameter selection algorithm")

    cfg.attack.parameter_index = parameter_index
    logging.info(f"Selected parameter index {parameter_index} for poisoning")
    return state_dict


@torch.no_grad()
def poison_model(cfg: DictConfig, model: nn.Module, reduction: str) -> None:
    """
    Check if the current step should be poisoned. If yes,
    the targeted parameter has its magnitude changed as
    following.
    1. + lr/batch_size * C if loss_reduction == "mean"
    2. + lr * C if loss_reduction == sum

    :param cfg: The hyperparameters of the experiment
    :param model: The attacked model
    :param step: The current step of optimization
    :return: Nothing, the poisoning is done in-place.
    """
    if cfg.logging_params.log_poisoning:
        logging.info(f"Poisoning model with a parameter index {cfg.attack.parameter_index}")

    parameter_index = cfg.attack.parameter_index
    layer_target_idx, *idx = find_index(parameter_index, param_shapes(model))
    for layer_idx, p in enumerate(model.parameters()):
        if layer_idx == layer_target_idx:
            if cfg.logging_params.log_poisoning:
                logging.info(f"Previous poisoned value: {p[tuple(idx)].item():.2f}")

            if reduction == "mean":
                magnitude = cfg.training_params.learning_rate * cfg.privacy.C / cfg.training_params.batch_size
            elif reduction == "sum":
                magnitude = cfg.training_params.learning_rate * cfg.privacy.C
            else:
                raise ValueError("Unknown loss reduction")

            p[tuple(idx)].sub_(magnitude)

            if cfg.logging_params.log_poisoning:
                logging.info(f"After poisoned value: {p[tuple(idx)].item():.2f}")
            return

    raise ValueError("Poisoning Failed")


def get_poison_attack_output(model: nn.Module, anomaly_gradient_feature: int) -> float:
    """
    Retrieve the value of the poisoned parameter.
    :param model: the attacked model
    :param anomaly_gradient_feature:
    :return:
    """
    layer_target_idx, *idx = find_index(anomaly_gradient_feature, param_shapes(model))

    for layer_idx, p in enumerate(model.parameters()):
        if layer_idx == layer_target_idx:
            return float(p[tuple(idx)])

    raise ValueError("No layer found")
