import logging
from typing import Tuple

import numpy as np
import torch
import torchvision
from torch.utils.data import DataLoader
from torchvision import transforms
from torchvision.transforms import Lambda

from empiricalprivacyestimation.data.gaussian_data_outlier import DataPoint


def get_cifar10(cfg) -> Tuple[DataLoader, DataLoader, DataPoint]:
    """Get the CIFAR10 dataset."""
    transform = transforms.Compose(
        [transforms.ToTensor(), transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))]
    )
    target_transform = Lambda(
        lambda y: torch.zeros(10, dtype=torch.float).scatter_(
            0, torch.tensor(y), value=1
        )
    )

    train = torchvision.datasets.CIFAR10(
        root=cfg.data_path,
        train=True,
        download=True,
        transform=transform,
        target_transform=target_transform,
    )

    test = torchvision.datasets.CIFAR10(
        root=cfg.data_path,
        train=False,
        download=True,
        transform=transform,
        target_transform=target_transform,
    )

    if cfg.attack.name == "loss":
        # pick a random data point and flip it's label
        attacked_sample_idx = np.random.randint(0, len(train))
        attacked_X = train.data[attacked_sample_idx]
        attacked_y = train.targets[attacked_sample_idx]
        logging.info("Attack sample_idx: %s", attacked_sample_idx)

        train.data = np.delete(train.data, attacked_sample_idx, axis=0)
        del train.targets[attacked_sample_idx]
        set_of_remainining_labels = list(set(list(range(10))) - {attacked_y})
        new_label = np.random.choice(set_of_remainining_labels)

        anomaly_point = (
            transform(attacked_X).unsqueeze(0),
            target_transform(new_label).unsqueeze(0),
        )
    elif cfg.attack.name == "poison":
        # ideal_canary
        anomaly_point = None
    elif cfg.attack.name == "canary_on_the_sphere":
        anomaly_point = None
    else:
        raise ValueError("Attack not supported.")

    trainloader = torch.utils.data.DataLoader(
        train,
        batch_size=cfg.training_params.batch_size,
        num_workers=8,
        shuffle=cfg.training_params.poisson_sampled_dataset,
        pin_memory=True,
    )

    testloader = torch.utils.data.DataLoader(
        test,
        batch_size=cfg.training_params.batch_size,
        shuffle=False,
        num_workers=8,
        pin_memory=True,
    )

    cfg.privacy.sample_rate = cfg.training_params.batch_size / len(train)
    return trainloader, testloader, anomaly_point
