from typing import Tuple

import torch
from omegaconf import DictConfig
from torch.utils.data import TensorDataset, DataLoader

from empiricalprivacyestimation.util.types import DataPoint


def get_gaussian_dataloaders(
    cfg: DictConfig,
    no_of_samples_per_class_train: int,
    no_of_samples_per_class_test: int,
    max_per_sample_grad_norm: float,
    batch_size: int,
    test_batch_size: int,
    number_of_regular_dimensions: int,
    data_variance: float,
    class_center_1: float,
    class_center_2: float,
) -> Tuple[DataLoader, DataLoader, DataPoint]:
    train_dataset, test_dataset, anomaly_point = gaussian_data(
        no_of_samples_per_class_train=no_of_samples_per_class_train - 1,
        no_of_samples_per_class_test=no_of_samples_per_class_test,
        dimensionality=number_of_regular_dimensions,
        sigma=data_variance,
        anomaly_dimension_outlier_value=max_per_sample_grad_norm,
        class_center_1=class_center_1,
        class_center_2=class_center_2,
    )
    size_of_dataset = len(train_dataset)

    if cfg.training_params.batch_size == "full_batch":
        cfg.training_params.batch_size = size_of_dataset
    else:
        if size_of_dataset // batch_size < size_of_dataset / batch_size:
            raise ValueError(
                "Please provide a batch size that divides the dataset evenly."
            )
    cfg.privacy.sample_rate = min(
        1.0, cfg.training_params.batch_size / len(train_dataset)
    )

    test_loader = torch.utils.data.DataLoader(
        test_dataset,
        batch_size=test_batch_size,
        shuffle=True,
        num_workers=0,
        pin_memory=True,
    )

    normal_loader = torch.utils.data.DataLoader(
        train_dataset,
        batch_size=cfg.training_params.batch_size,
        num_workers=0,
        pin_memory=True,
    )

    return normal_loader, test_loader, anomaly_point


def gaussian_data(
    no_of_samples_per_class_train: int,
    no_of_samples_per_class_test: int,
    dimensionality: int,
    class_center_1: float,
    class_center_2: float,
    sigma: float,
    anomaly_dimension_outlier_value: int,
) -> Tuple[TensorDataset, TensorDataset, DataPoint]:
    no_classes = 2
    # two labels in our case: [0, 1]
    train_class = torch.arange(0, no_classes)

    # generate the selected number of labels for the train set
    train_y = train_class.repeat_interleave(no_of_samples_per_class_train)

    # randomly pick the centers for our classes according to the dimensionality
    class_centers = torch.ones(no_classes, dimensionality)
    class_centers[0, :] = class_centers[0, :] * class_center_1
    class_centers[1, :] = class_centers[1, :] * class_center_2

    # generate the samples under some noise around the class centers
    train_x = class_centers[train_y] + sigma * torch.randn(
        no_of_samples_per_class_train * no_classes, dimensionality
    )

    # add a dimension of 0 - the abnormal dimension
    train_x = torch.cat(
        (train_x, torch.zeros((no_of_samples_per_class_train * no_classes, 1))), dim=1
    )

    # repeat the process for the test set
    test_y = train_class.repeat_interleave(no_of_samples_per_class_test)
    test_x = class_centers[test_y] + sigma * torch.randn(
        no_of_samples_per_class_test * no_classes, dimensionality
    )
    test_x = torch.cat(
        (test_x, torch.zeros((no_of_samples_per_class_test * no_classes, 1))), dim=1
    )

    # generate the outlier point by taking the mean of the existing class centers
    outlier_point_features = torch.zeros(dimensionality)
    outlier_point_features = torch.cat(
        (
            outlier_point_features,
            torch.tensor(anomaly_dimension_outlier_value).unsqueeze(0),
        )
    )

    # add a non-0 value for the last dimension
    anomalous_train_x = torch.cat((train_x, outlier_point_features.unsqueeze(0)), dim=0)

    # pick a random label for this sample
    anomalous_train_y = torch.cat((train_y, torch.tensor(0).unsqueeze(0)))

    # the dataset without the anomalous point
    data = TensorDataset(train_x, train_y.unsqueeze(1).float())
    anomalous_x, anomalous_y = (
        anomalous_train_x[-1].unsqueeze(0),
        anomalous_train_y[-1].reshape(1, 1).float(),
    )
    test_data = TensorDataset(test_x, test_y.unsqueeze(1).float())

    return data, test_data, (anomalous_x, anomalous_y)
