import logging

import numpy as np
import torch
from omegaconf import DictConfig
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from torch.utils.data import TensorDataset


def get_housing_data(cfg: DictConfig):
    data = np.load(f"{cfg.data_path}/houses.npy")
    scaler = StandardScaler()
    data = scaler.fit(data).transform(data)
    X = data[:, :-1]
    y = data[:, -1]

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, random_state=42
    )

    number_of_labels = len(set(y_train))
    # scaler = StandardScaler()
    # X_train = scaler.fit_transform(X_train)
    # X_test = scaler.transform(X_test)

    if cfg.attack.name == "loss":
        # pick a random data point and flip it's label
        attacked_sample_idx = np.random.randint(0, len(X_train))
        attacked_X = X_train[attacked_sample_idx]
        attacked_y = y_train[attacked_sample_idx]
        logging.info("Attack sample_idx: %s", attacked_sample_idx)

        X_train = np.delete(X_train, attacked_sample_idx, axis=0)
        y_train = np.delete(y_train, attacked_sample_idx)
        set_of_remainining_labels = list(set(list(range(number_of_labels))) - {attacked_y})
        new_label = np.random.choice(set_of_remainining_labels)

        anomaly_point = (
            torch.tensor(attacked_X).unsqueeze(0).float(),
            torch.tensor(new_label).unsqueeze(0).long(),
        )
    else:
        anomaly_point = None

    train_dataset = TensorDataset(
        torch.from_numpy(X_train).float(), torch.from_numpy(y_train).long()
    )
    test_dataset = TensorDataset(
        torch.from_numpy(X_test).float(), torch.from_numpy(y_test).long()
    )

    if hasattr(cfg.training_params, "shuffle"):
        shuffle = cfg.training_params.shuffle
    else:
        shuffle = False
    size_of_dataset = len(train_dataset)
    if cfg.training_params.batch_size == "full_batch":
        cfg.training_params.batch_size = size_of_dataset
    else:
        if size_of_dataset // cfg.training_params.batch_size < size_of_dataset / cfg.training_params.batch_size:
            raise ValueError(
                "Please provide a batch size that divides the dataset evenly."
            )
        
    normal_loader = torch.utils.data.DataLoader(
        train_dataset,
        batch_size=cfg.training_params.batch_size,
        shuffle=shuffle,
        num_workers=0,
        pin_memory=True,
    )

    test_loader = torch.utils.data.DataLoader(
        test_dataset,
        batch_size=cfg.training_params.batch_size,
        shuffle=True,
        num_workers=0,
        pin_memory=True,
    )

    return normal_loader, test_loader, anomaly_point