import logging
from typing import Tuple

import numpy as np
import torch
from omegaconf import DictConfig
from sklearn.datasets import load_iris
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from torch.utils.data import DataLoader, TensorDataset

from empiricalprivacyestimation.util.types import DataPoint


def get_iris(cfg: DictConfig) ->  Tuple[DataLoader, DataLoader, DataPoint]:


    iris = load_iris()
    X = iris.data
    y = iris.target

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, random_state=42
    )

    scaler = StandardScaler()
    X_train = scaler.fit_transform(X_train)
    X_test = scaler.transform(X_test)


    if cfg.attack.name == "loss":
        # pick a random data point and flip it's label
        attacked_sample_idx = np.random.randint(0, len(X_train))
        attacked_X = X_train[attacked_sample_idx]
        attacked_y = y_train[attacked_sample_idx]
        logging.info("Attack sample_idx: %s", attacked_sample_idx)

        X_train = np.delete(X_train, attacked_sample_idx, axis=0)
        del y_train[attacked_sample_idx]
        set_of_remainining_labels = list(set(list(range(10))) - {attacked_y})
        new_label = np.random.choice(set_of_remainining_labels)

        anomaly_point = (
            attacked_X.unsqueeze(0),
            new_label.unsqueeze(0),
        )
    else:
        anomaly_point = None

    train_dataset = TensorDataset(
        torch.from_numpy(X_train).float(), torch.from_numpy(y_train).long()
    )
    test_dataset = TensorDataset(
        torch.from_numpy(X_test).float(), torch.from_numpy(y_test).long()
    )


    size_of_dataset = len(train_dataset)

    if cfg.training_params.batch_size == "full_batch":
        cfg.training_params.batch_size = size_of_dataset
    else:
        if size_of_dataset // cfg.training_params.batch_size < size_of_dataset / cfg.training_params.batch_size:
            raise ValueError(
                "Please provide a batch size that divides the dataset evenly."
            )
    cfg.privacy.sample_rate = min(
        1.0, cfg.training_params.batch_size / len(train_dataset)
    )

    normal_loader = torch.utils.data.DataLoader(
        train_dataset,
        batch_size=cfg.training_params.batch_size,
        num_workers=0,
        pin_memory=True,
        shuffle=False
    )

    test_loader = torch.utils.data.DataLoader(
        test_dataset,
        batch_size=cfg.training_params.batch_size,
        shuffle=False,
        num_workers=0,
        pin_memory=True,
    )



    return normal_loader, test_loader, anomaly_point