import logging
from typing import Tuple

import numpy as np
from joblib import Parallel, delayed
from sklearn import metrics
from sklearn.metrics import confusion_matrix
from statsmodels.stats import proportion

from .clopper_pearson import get_bounds_clopper
from .gdp_cp import get_bounds as get_bouds_gdp


#
# def betainc_inv(a, b, p, method='bisect'):
#     """
#     Compute the inverse of the regularized incomplete beta function.
#
#     Parameters:
#     - a, b: Shape parameters of the beta distribution.
#     - p: Cumulative probability.
#     - method: Root-finding method ('bisect', 'newton', 'secant', etc.)
#
#     Returns:
#     - x: Value such that I_x(a, b) = p
#     """
#     # Ensure p is within (0, 1)
#     if p <= 0:
#         return mp.mpf(0)
#     if p >= 1:
#         return mp.mpf(1)
#
#     # Define the function whose root we want to find
#     def f(x):
#         return betainc(a, b, 0, x, regularized=True) - p
#
#     # Initial guesses for x
#     x0 = mp.mpf(0)
#     x1 = mp.mpf(1)
#
#     # Use findroot to solve f(x) = 0
#     x = findroot(f, (x0, x1), method=method, solver='bisect', tol=mp.eps)
#     return x
#
# def clopper_pearson_confint_high_precision(k, n, alpha=0.05, precision=50):
#     """
#     Compute the Clopper-Pearson exact confidence interval for a proportion using high-precision arithmetic.
#
#     Parameters:
#     - k: Number of successes (integer)
#     - n: Total number of trials (integer)
#     - alpha: Significance level (default 0.05 for a 95% confidence interval)
#     - precision: Number of decimal places for high-precision arithmetic
#
#     Returns:
#     - (lower_bound, upper_bound): The confidence interval
#     """
#     # Set the desired precision
#     mp.dps = precision
#
#     k = int(k)
#     n = int(n)
#
#     if k < 0 or n <= 0 or k > n:
#         raise ValueError("Invalid values: ensure 0 <= k <= n and n > 0.")
#
#     alpha = mp.mpf(alpha)
#     k = mp.mpf(k)
#     n = mp.mpf(n)
#
#     if k == 0:
#         lower_bound = mp.mpf(0)
#         upper_bound = mp.powm1(alpha / 2, mp.mpf(1) / n, mp.mpf(1))
#     elif k == n:
#         lower_bound = mp.pow(alpha / 2, mp.mpf(1) / n)
#         upper_bound = mp.mpf(1)
#     else:
#         # Compute the lower and upper bounds using betaincinv
#         lower_bound = betainc_inv(k, n - k + 1, alpha / 2)
#         upper_bound = betainc_inv(k + 1, n - k, 1 - alpha / 2)
#
#     # Convert back to floats
#     lower_bound = float(lower_bound)
#     upper_bound = float(upper_bound)
#
#     return lower_bound, upper_bound


def get_privacy_profile(
    cfg, outputs: np.ndarray, gt: np.ndarray, alpha=0.05, method="beta", threshold=None
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    def alpha_beta_single(threshold: float) -> Tuple[float, float, float]:
        predictions = outputs > threshold
        tn, fp, fn, tp = confusion_matrix(gt, predictions).ravel()

        sum = tn + tp + fp + fn

        if (tn + tp) / sum < (fp + fn) / sum:
            tn, fp = fp, tn
            fn, tp = tp, fn

        _, p1_ub = proportion.proportion_confint(fn, tp + fn, alpha, method=method)
        _, p0_ub = proportion.proportion_confint(fp, tn + fp, alpha, method=method)

        # _, p1_ub = clopper_pearson_confint_high_precision(fn, tp + fn, alpha)
        # _, p0_ub = clopper_pearson_confint_high_precision(fp, tn + fp, alpha)

        # _, p1_ub_v0 = clopper_pearson_confint(fn, tp + fn, alpha)
        # _, p0_ub_v0 = clopper_pearson_confint(fp, tn + fp, alpha)
        # print(p0_ub, p1_ub, p0_ub_v0, p1_ub_v0)

        if p0_ub < cfg.privacy.delta or \
                p1_ub < cfg.privacy.delta or\
                p0_ub > 1 - cfg.privacy.delta or \
                p1_ub > 1 - cfg.privacy.delta or \
                p0_ub + p1_ub > 1:
            return 0., 0., 0., 0.

        try:
            eps_lb_gdp = get_bouds_gdp(fnr=p0_ub, fpr=p1_ub, delta=cfg.privacy.delta)
        except Exception as e:
            logging.info(str(e))
            eps_lb_gdp = 0

        try:
            eps_lb_cp = get_bounds_clopper(fnr=p0_ub, fpr=p1_ub, delta=cfg.privacy.delta)
        except Exception as e:
            logging.info(str(e))
            eps_lb_cp = 0

        return p0_ub, p1_ub, eps_lb_cp, eps_lb_gdp

    if threshold:
        alphas_ub, betas_ub, epsilon_cp_lbs, epsilon_gdp_lbs = alpha_beta_single(threshold)
        return alphas_ub, betas_ub, epsilon_cp_lbs, epsilon_gdp_lbs

    _, _, thresholds = metrics.roc_curve(gt, outputs)
    result = Parallel(n_jobs=8)(
        delayed(alpha_beta_single)(threshold) for threshold in thresholds
    )
    alphas_ub = np.array([x[0] for x in result])
    betas_ub = np.array([x[1] for x in result])
    epsilon_cp_lbs = np.array([x[2] for x in result])
    epsilon_gdp_lbs = np.array([x[3] for x in result])

    if np.all(epsilon_gdp_lbs == 0):
        logging.info("No valid data")
        return alphas_ub, betas_ub, epsilon_cp_lbs, epsilon_gdp_lbs

    mask_1 = np.logical_and(alphas_ub > cfg.privacy.delta, alphas_ub < 1 - cfg.privacy.delta)
    mask_2 = np.logical_and(betas_ub > cfg.privacy.delta, betas_ub < 1 - cfg.privacy.delta)
    mask = np.logical_and(mask_1, mask_2)
    alphas_ub = alphas_ub[mask]
    betas_ub = betas_ub[mask]
    epsilon_cp_lbs = epsilon_cp_lbs[mask]
    epsilon_gdp_lbs = epsilon_gdp_lbs[mask]
    return alphas_ub, betas_ub, epsilon_cp_lbs, epsilon_gdp_lbs