import logging
from copy import deepcopy

import torch


class BoundedSetOptimization:
    def __init__(self, cfg, initial_state_dict):
        self.bounded_set_radius = cfg.param_regularization.bounded_set_radius
        self.initial_state_dict = initial_state_dict

    def __call__(self, module):
        if hasattr(module, "weight"):
            with torch.no_grad():
                logging.info("Performing clipping in constrained optimization")
                logging.info(f"Initial params: {module.weight}")
                key_name = f"{module._get_name().lower()}.weight"
                w_init = self.initial_state_dict[key_name]

                w = module.weight.data
                w = (w - w_init).clamp(
                    -self.bounded_set_radius, self.bounded_set_radius
                ) + w_init
                module.weight.data = w
                logging.info(f"After clamping: {module.weight}")


class LogisticRegression(torch.nn.Module):
    def __init__(self, cfg):
        super(LogisticRegression, self).__init__()
        self.linear = torch.nn.Linear(
            cfg.data.number_of_regular_dimensions + 1, 1, bias=False
        )

        with torch.no_grad():
            self.linear.weight[-1][-1].zero_().add_(100)

        if cfg.param_regularization.name == "bounded_set":
            self.projection = BoundedSetOptimization(
                cfg=cfg,
                initial_state_dict=deepcopy(self.state_dict()),
            )
        else:
            self.projection = lambda x: x

    def forward(self, x, stable: bool = False):
        if stable:
            return self.linear(x)
        else:
            return torch.sigmoid(self.linear(x))

    @property
    def anomalous_dimension(self):
        return list(self.parameters())[0].squeeze()[-1]

    @property
    def grad(self):
        return list(self.parameters())[0].grad
