import logging
from typing import List, Tuple

from torch.nn import Module
import torch


def flat_grad(model: Module) -> torch.Tensor:
    return torch.cat([p.grad.flatten() for p in model.parameters() if p.requires_grad])


def flat_params(model: Module) -> torch.Tensor:
    return torch.cat([p.flatten() for p in model.parameters() if p.requires_grad])


def no_params(model: Module) -> int:
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def param_shapes(model: Module) -> Tuple[Tuple[int, ...], ...]:
    param_shapes = []
    for p in model.parameters():
        param_shapes.append(p.shape)
    return tuple(param_shapes)


class BoundedSetProjection:
    def __init__(self, cfg):
        """
        Note: We assume initialization in 0.
        :param cfg:
        """
        self.bounded_set_radius = cfg.param_regularization.bounded_set_radius

    def __call__(self, module):
        if hasattr(module, "weight"):
            with torch.no_grad():
                logging.info("Performing clipping in constrained optimization")
                logging.info(f"Initial params: {module.weight}")

                w = module.weight.data
                w = w.clamp(-self.bounded_set_radius, self.bounded_set_radius)
                module.weight.data = w
                logging.info(f"After clamping: {module.weight}")
