from typing import Callable, Optional, Tuple, Dict

import numpy as np
import torch
from omegaconf import DictConfig
from torch.utils.data import DataLoader

from empiricalprivacyestimation.attacks import attack_setup
from empiricalprivacyestimation.training.accounting import (
    get_privacy_guarantees,
)
from empiricalprivacyestimation.training.cpu_multithread import (
    cpu_multithread_attack_samples_pipeline,
)
from empiricalprivacyestimation.training.gpu_multithread import (
    gpu_multithread_attack_samples_pipeline,
)
from empiricalprivacyestimation.training.singlethread import (
    singledevice_attack_samples_pipeline,
)


def get_training_regime(cfg: DictConfig) -> Callable:
    if cfg.training_params.device == "cpu":
        if cfg.training_params.multidevice:
            return cpu_multithread_attack_samples_pipeline
        else:
            return singledevice_attack_samples_pipeline

    if cfg.training_params.device.startswith("cuda"):
        if cfg.training_params.multidevice:
            return gpu_multithread_attack_samples_pipeline
        else:
            return singledevice_attack_samples_pipeline

    raise ValueError(
        f"Device {cfg.training_params.device} multidevice: {cfg.training_params.multidevice} not supported"
    )


def get_attack_scores(
    cfg: DictConfig,
    normal_loader: DataLoader,
    test_loader: DataLoader,
    attack_point: Tuple[torch.Tensor, torch.Tensor],
    state_dict: Optional[Dict] = None,
) -> Tuple[Dict, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    # shape is (number_of_attack_runs, 2), the first dimension is the gt, the second dimension is the attacker guess
    state_dict = attack_setup(cfg, normal_loader, test_loader, state_dict=state_dict)
    if cfg.privacy.compute_epsilons:
        epsilons = get_privacy_guarantees(cfg)
    else:
        epsilons = np.array([])

    attack_samples_pipeline = get_training_regime(cfg)
    attack_outputs, gt, test_scores, sensitivity = attack_samples_pipeline(
        cfg=cfg,
        normal_loader=normal_loader,
        test_loader=test_loader,
        state_dict=state_dict,
        attack_point=attack_point,
    )

    attack_outputs = np.array(attack_outputs)
    gt = np.array(gt)
    return state_dict, attack_outputs, gt, epsilons, test_scores, sensitivity
