import logging
import time
from copy import deepcopy
from typing import Dict, Tuple

import numpy as np
import ray
from omegaconf import DictConfig
from psutil import cpu_count
from torch.utils.data import DataLoader

from empiricalprivacyestimation.training.singlethread import (
    singledevice_attack_samples_pipeline,
)
from empiricalprivacyestimation.util.types import DataPoint


def cpu_multithread_attack_samples_pipeline(
    cfg: DictConfig,
    normal_loader: DataLoader,
    test_loader: DataLoader,
    state_dict: Dict,
    attack_point: DataPoint,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    start = time.time()
    ray.init()
    logging.info("Starting multithreaded CPU samples pipeline")
    no_of_threads = cpu_count(logical=True) // 2
    logging.info(f"Number of threads: {no_of_threads}")
    load_share = cfg.empirical_epsilon.attack_samples // no_of_threads
    logging.info(f"Load share: {load_share}")
    worker_cfg = deepcopy(cfg)
    worker_cfg.empirical_epsilon.attack_samples = load_share
    normal_loader_ptr = ray.put(normal_loader)
    test_loader_ptr = ray.put(test_loader)
    state_dict_ptr = ray.put(state_dict)
    attack_point_ptr = ray.put(attack_point)

    taskfunc = ray.remote(singledevice_attack_samples_pipeline)

    tasks = [taskfunc.remote(
            cfg=worker_cfg,
            normal_loader=normal_loader_ptr,
            test_loader=test_loader_ptr,
            state_dict=state_dict_ptr,
            attack_point=attack_point_ptr,
        )
        for _ in range(no_of_threads)
    ]
    results = ray.get(tasks)
    end = time.time()
    logging.info(f"Multithreaded CPU samples pipeline took {end - start} seconds")

    attack_outputs = np.concatenate(
        list(results_per_thread[0] for results_per_thread in results)
    )
    gt = np.concatenate(list(results_per_thread[1] for results_per_thread in results))
    test_scores = np.concatenate(
        list(results_per_thread[2] for results_per_thread in results)
    )
    sensitivities = np.concatenate(
        list(results_per_thread[3] for results_per_thread in results)
    )

    ray.shutdown()
    return attack_outputs, gt, test_scores, sensitivities
