import logging
import time
from copy import deepcopy
from typing import Dict, Tuple

import numpy as np
import ray
import torch
from omegaconf import DictConfig
from psutil import cpu_count
from torch.utils.data import DataLoader

from empiricalprivacyestimation.training.singlethread import (
    singledevice_attack_samples_pipeline,
)
from empiricalprivacyestimation.util.types import DataPoint


def gpu_multithread_attack_samples_pipeline(
    cfg: DictConfig,
    normal_loader: DataLoader,
    test_loader: DataLoader,
    state_dict: Dict,
    attack_point: DataPoint,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:

    start = time.time()
    logging.info("Initializing ray")
    gpus = [torch.cuda.device(i) for i in range(torch.cuda.device_count())]
    if len(gpus) == 1:
        logging.info("No parallelism needed, running on 1 GPU.")
        # no need of parallelism
        return singledevice_attack_samples_pipeline(
            cfg=cfg,
            normal_loader=normal_loader,
            test_loader=test_loader,
            state_dict=state_dict,
            attack_point=attack_point,
        )

    logging.info(f"Number of GPUs: {len(gpus)}")
    logging.info(f"Number of CPUs: {cpu_count(logical=True)}")
    ray.init(
        num_gpus=len(gpus),
        num_cpus=cpu_count(logical=True),
        ignore_reinit_error=True,
        include_dashboard=False
    )

    no_of_workers = len(gpus)
    cpu_share_per_gpu = cpu_count(logical=True) // len(gpus)
    logging.info(
        f"Starting multithreaded GPU samples pipeline - worker resources: GPU:1 CPU: {cpu_share_per_gpu}"
    )
    logging.info(f"Number of threads: {no_of_workers}")

    load_share = cfg.empirical_epsilon.attack_samples // no_of_workers
    logging.info(f"Load share: {load_share}")

    taskfunc = ray.remote(
        num_gpus=1
      )(singledevice_attack_samples_pipeline)

    tasks = []
    for worker_id in range(no_of_workers):
        worker_cfg = deepcopy(cfg)
        worker_cfg.empirical_epsilon.attack_samples = load_share
        worker_cfg.training_params.device = f"cuda"
        logging.info(f"Worker {worker_id}")
        task = taskfunc.remote(
            cfg=worker_cfg,
            normal_loader=normal_loader,
            test_loader=test_loader,
            state_dict=state_dict,
            attack_point=attack_point,
        )
        tasks.append(task)

    results = ray.get(tasks)
    end = time.time()
    logging.info(f"Multithreaded GPU samples pipeline took {end - start} seconds")

    attack_outputs = np.concatenate(
        list(results_per_thread[0] for results_per_thread in results)
    )
    gt = np.concatenate(list(results_per_thread[1] for results_per_thread in results))
    test_scores = np.concatenate(
        list(results_per_thread[2] for results_per_thread in results)
    )
    sensitivities = np.concatenate(
        list(results_per_thread[3] for results_per_thread in results)
    )

    # Wait until placement group is killed.
    time.sleep(1)
    ray.shutdown()

    return attack_outputs, gt, test_scores, sensitivities
