import logging
from typing import Dict, Tuple

import numpy as np
import torch.cuda
from omegaconf import DictConfig
from torch.utils.data import DataLoader
from tqdm import tqdm

from empiricalprivacyestimation.training.training_routines import get_single_score
from empiricalprivacyestimation.util.types import DataPoint


def singledevice_attack_samples_pipeline(
    cfg: DictConfig,
    normal_loader: DataLoader,
    test_loader: DataLoader,
    state_dict: Dict,
    attack_point: DataPoint,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    if torch.cuda.is_available():
        gpus = [torch.cuda.device(i) for i in range(torch.cuda.device_count())]
        logging.info(f"Number of GPUs: {len(gpus)}")
        logging.info(f"GPUs: {gpus}")
        logging.info(f"Number of CUDA devices: {torch.cuda.device_count()}")
        logging.info(f"CUDA devices: {[torch.cuda.get_device_name(i) for i in range(torch.cuda.device_count())]}")
    else:
        logging.info("No CUDA device available")
    gt = []
    attack_outputs = []
    test_scores = []
    sensitivities = []
    for i in tqdm(range(cfg.empirical_epsilon.attack_samples)):
        (attack_run, b, test_score, sensitivity) = get_single_score(
            cfg=cfg,
            train_loader=normal_loader,
            test_loader=test_loader,
            state_dict=state_dict,
            attack_point=attack_point,
        )
        attack_outputs.append(attack_run)
        gt.append(b)
        test_scores.append(test_score)
        sensitivities.append(sensitivity)
        logging.info(
            f"Sample run: {i} - without anomaly: {bool(b)} - sigma: {cfg.privacy.sigma} - grad norm: {cfg.privacy.C}"
        )

    return (
        np.array(attack_outputs),
        np.array(gt),
        np.array(test_scores),
        np.array(sensitivities),
    )
