import gc
import logging
from typing import Dict, Tuple, List

import numpy
import numpy as np
import torch
from omegaconf import DictConfig
from opacus import PrivacyEngine
from torch import nn
from torch.utils.data import DataLoader
from torchvision import models, transforms

from empiricalprivacyestimation.attacks import get_attack_score
from empiricalprivacyestimation.attacks.canary_on_the_sphere import embed_canary_in_model
from empiricalprivacyestimation.attacks.poison import poison_model
from empiricalprivacyestimation.attacks.util import insert_attacked_sample
from empiricalprivacyestimation.models import get_model
from empiricalprivacyestimation.util.loss_fn import get_loss_fn
from empiricalprivacyestimation.util.measure_sample_sensitivity import measure_sample_sensitivity
from empiricalprivacyestimation.util.optim import get_optim
from empiricalprivacyestimation.util.types import DataPoint


def get_single_score(
    cfg: DictConfig,
    train_loader: DataLoader,
    test_loader: DataLoader,
    state_dict: Dict,
    attack_point: DataPoint,
) -> Tuple[np.ndarray, int, float, np.ndarray]:
    # flip a coin to decide if we are using D or D'
    # b = 1 - the dataset without the anomaly (D)
    # b = 0 - the dataset with the anomaly (D')
    b = np.random.choice([0, 1])
    attacked_steps = pregenerate_attacked_steps(cfg, b)
    model, attack_run, test_score, sensitivity = train_model(
        cfg=cfg,
        train_loader=train_loader,
        test_loader=test_loader,
        state_dict=state_dict,
        attacked_steps=attacked_steps,
        attack_point=attack_point,
        b=b
    )

    # cleanup
    model.cpu()
    del model
    gc.collect()
    torch.cuda.empty_cache()

    return attack_run, b, test_score, sensitivity


def pregenerate_attacked_steps(cfg: DictConfig, b) -> np.ndarray:
    attacked_steps = np.zeros(cfg.training_params.total_no_of_steps + 1, dtype=np.bool_)

    if b:
        return attacked_steps

    target = cfg.accounting.canary_sampling
    if target.startswith("every"):
        k = int(target.split("@")[1])
        indexes = [k * i + 1 for i in range(cfg.training_params.total_no_of_steps // k)]
        attacked_steps[indexes] = 1
    else:
        raise ValueError("Unsupported canary insertion scheme")

    return attacked_steps


def train_model(
    cfg: DictConfig,
    train_loader: DataLoader,
    test_loader: DataLoader,
    state_dict: Dict,
    attack_point: DataPoint,
    attacked_steps: np.ndarray,
    b: bool
) -> Tuple[nn.Module, np.ndarray, List[float], np.ndarray]:
    model = get_model(cfg)
    model.load_state_dict(state_dict)

    if cfg.training_params.finetune:
        logging.info("loading weights for alexnet")
        # resize all datasets to 224x224
        transform = transforms.Compose(
            [
                transforms.Resize((224, 224)),
                transforms.ToTensor(),
            ]
        )
        train_loader.dataset.transform = transform
        alexnet = models.alexnet(pretrained=False)
        alexnet.load_state_dict(torch.load(cfg.weight_path))
        alexnet.eval()
        # send to device
        alexnet.to(cfg.training_params.device)
        del alexnet.classifier[6]

    # if poisson sampling, sample with probability q when to insert the data point
    if (not b) and hasattr(cfg.privacy, "sample_canary") and cfg.privacy.sample_canary:
        # flip a coin with probability sampling rate
        q = cfg.privacy.sample_rate
        attacked_steps = np.random.choice([0, 1], size=cfg.training_params.total_no_of_steps + 1, p=[1 - q, q])
        logging.error(f"get new attacked steps: {attacked_steps}")

    init_model = get_model(cfg)
    if cfg.training_params.fixed_init:
        init_model.load_state_dict(state_dict)

    optimizer = get_optim(cfg, model)

    if not cfg.privacy.disable_dp:
        privacy_engine = PrivacyEngine(secure_mode=False)
        _, reduction = get_loss_fn(cfg)
        model, optimizer, poisson_train_loader = privacy_engine.make_private(
            module=model,
            optimizer=optimizer,
            data_loader=train_loader,
            noise_multiplier=cfg.privacy.sigma,
            max_grad_norm=cfg.privacy.C,
            loss_reduction=reduction,
        )

        if cfg.training_params.poisson_sampled_dataset:
            train_loader = poisson_train_loader

    attack_scores = numpy.zeros(cfg.training_params.total_no_of_steps + 1)
    sensitivity_scores = numpy.zeros(cfg.training_params.total_no_of_steps + 1)
    test_scores = []


    attack_scores[0] = get_attack_score(cfg=cfg, model=model, attack_point=attack_point, initialization=init_model)

    if cfg.attack.name == "loss":
        sensitivity_scores[0] = measure_sample_sensitivity(cfg, attack_point, model)
    else:
        sensitivity_scores[0] = cfg.privacy.C

    dataloader_iterator = iter(train_loader)
    criterion, reduction = get_loss_fn(cfg)
    device = torch.device(cfg.training_params.device)

    for curr_step in range(1, cfg.training_params.total_no_of_steps + 1):
        optimizer.zero_grad()
        model.train()

        try:
            data, target = next(dataloader_iterator)
        except:
            dataloader_iterator = iter(train_loader)
            data, target = next(dataloader_iterator)

        is_attacked_step = attacked_steps[curr_step]

        if is_attacked_step and (cfg.attack.name == "loss"):
            data, target = insert_attacked_sample(cfg, data, target, attack_point)

        data, target = data.to(device), target.to(
            device
        )
        if cfg.training_params.finetune:
            with torch.no_grad():
                data = alexnet(data)

        if cfg.model.name == "LogisticRegression" or cfg.model.name == "FCNN":
            output = model(data, stable=True)
        else:
            output = model(data)

        loss = criterion(output, target)
        loss.backward()

        if cfg.logging_params.log_attack_score:
            logging.info(f"Attack score before step {curr_step} - {attack_scores[curr_step - 1]}")

        optimizer.step()
        if is_attacked_step and (cfg.attack.name == "poison"):
            poison_model(cfg, model, reduction=reduction)

        if is_attacked_step and cfg.attack.name == "canary_on_the_sphere":
            embed_canary_in_model(cfg, model, reduction=reduction)

        score = get_attack_score(cfg=cfg, model=model, attack_point=attack_point, initialization=init_model)

        attack_scores[curr_step] = score

        if cfg.logging_params.log_attack_score:
            logging.info(f"Attack score after step {curr_step} - {attack_scores[curr_step]}")

        if cfg.attack.name == "loss" and cfg.logging_params.log_sensitivity:
             optimizer.zero_grad()
             sensitivity = measure_sample_sensitivity(cfg, attack_point, model)
        else:
             sensitivity = cfg.privacy.C

        sensitivity_scores[curr_step] = sensitivity

        if hasattr(cfg.logging_params, "track_performance") and cfg.logging_params.track_performance:
            if curr_step % 100 == 0:
                with torch.no_grad():
                    test_score = test(cfg, model, test_loader)
                    test_scores.append(test_score)

                logging.info(f"Step {curr_step} - test score: {test_score}")

    return model, numpy.array(attack_scores), test_scores, sensitivity_scores


def test(cfg: DictConfig, model: nn.Module, test_loader: DataLoader) -> float:
    device = torch.device(cfg.training_params.device)
    model.eval()
    test_loss = 0
    correct = 0
    criterion, reduction = get_loss_fn(cfg)

    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            test_loss += criterion(output, target).item()  # sum up batch loss
            if output.shape[-1] == 1:
                pred = torch.round(output)
                correct += pred.eq(target).sum().item()
            else:
                predicted = torch.argmax(output.data, 1)
                if not (cfg.model.name == "LogisticRegressionIris" or cfg.model.name == "fcnn"):
                    target = torch.argmax(target, 1)
                correct += (predicted == target).sum().item()

    return correct / len(test_loader.dataset)
